#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 26 14:02:49 2018
Optimisation of the cost function using a differential evolution solver.
c
@author: Jerome Stenger
"""
from __future__ import division, print_function

import os
import sys
import time
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE,SIG_DFL) 

from math import *
import numpy as np
import matplotlib.pyplot as plt

from Model import my_model, Cost_Function

#%%

# =============================================================================
# =============================================================================
############################## SOLVER PARAMETERS ##############################
# =============================================================================
# =============================================================================

MINMAX = 1  ## NOTE: sup = maximize = -1; inf = minimize = 1
npop = 40 # size of the overall population
maxiter = 200 # maximum number of population mutation
maxfun = 1e+6 
convergence_tol = 1e-6; ngen = 20 # calculation stops with precision equal to tolerance after ngen iteration
crossover = 0.01 # probability of mutation 
percent_change = 0.01 # Scaling factor
#Xinit=[160,3580,12.55,47.45,50,50,54.5,54.5]

def optimizeDE(cost,xmin,xmax):  #solver for the second case
	from mystic.solvers import DifferentialEvolutionSolver2
	from mystic.termination import ChangeOverGeneration as COG
	from mystic.strategy import Best1Exp
	from mystic.monitors import VerboseMonitor, Monitor
	
	stepmon = VerboseMonitor(5) #step for showing live update
	evalmon = Monitor()
	ndim=len(xmin)
	
	solver = DifferentialEvolutionSolver2(ndim, npop)
	solver.SetRandomInitialPoints(xmin, xmax)
	#solver.SetInitialPoints(Xinit,radius=0.05) # Generate random points within [-radius*x0, radius*x0]
	solver.SetStrictRanges(xmin, xmax)
	solver.SetEvaluationLimits(maxiter,maxfun)
	solver.SetEvaluationMonitor(evalmon)
	solver.SetGenerationMonitor(stepmon)
	tol=convergence_tol
	solver.Solve(cost,termination=COG(tolerance=tol, generations=ngen),strategy=Best1Exp,CrossProbability=crossover,ScalingFactor=percent_change) 
	solved = solver.bestSolution
	func_max = MINMAX * solver.bestEnergy
	func_evals = solver.evaluations
	return solved, func_max, func_evals



#%%
# =============================================================================
# =============================================================================
############################ CALCULATION RUN ##################################
# =============================================================================
# =============================================================================

# name of the file
name='Differential Evolution'

# threshold bounds
threshold_max = 5
threshold_min = 3
step = 0.1

lower=[160]
upper=[3580]

mean=list((np.asarray(upper)+np.asarray(lower))/2.) # You can put any value want in a list



#rewritte the function to be implemented in the solver
def costDE(x0):
	global threshold, lower, upper, mean
	x=[[x0[2*t],x0[2*t+1]] for t in range(int(len(x0)/2))] #Concatenation of the vector per variable
	return Cost_Function(x, threshold=threshold, lower=lower,upper=upper,mean=mean)


res=[]
threshold=threshold_max
xmin=[]
xmax=[]
for i in range(len(lower)):
	xmin.append(lower[i]);xmin.append(mean[i])
	xmax.append(mean[i]);xmax.append(upper[i])
x=[]
y=[]
plt.rc('text', usetex=True)
plt.rc('font', family='serif')
plt.rc('xtick', labelsize=20) 
plt.rc('ytick', labelsize=20) 
plt.ylabel(r'$\inf_{\mu\in\mathcal{A}} \; F_{\mu}(h)$',fontsize=24)
plt.xlabel('h',fontsize=24)
plt.title(name,fontsize=24)
plt.show()
axes = plt.gca()
axes.set_xlim(threshold_min,threshold_max) # x-axis bounds
axes.set_ylim(0,1) # y-axis bounds
line, = axes.plot(x, y, 'r-')

start = time.time()
while threshold>=threshold_min:
	solved, func_max, func_evals = optimizeDE(costDE,xmin,xmax)
	x.append(threshold)
	y.append(func_max)
	line.set_xdata(x) # add new points to the figure
	line.set_ydata(y) 
	plt.draw() #update the figure
	plt.pause(1e-17)
	threshold+=-step
plt.show()
end = time.time()
print('time elapsed :',end - start)
time.sleep(0.5)

#%%
# =============================================================================
# =============================================================================
################### RESULT VISUALISATION AND EXPORTATION#######################
# =============================================================================
# =============================================================================
plt.close('all')

#Print and saved fig to be printed with latex 
plt.figure(figsize=(12,7))
plt.rc('text', usetex=True)
plt.rc('font', family='serif')
plt.rc('xtick', labelsize=20) 
plt.rc('ytick', labelsize=20) 
plt.ylabel(r'$\inf_{\mu\in\mathcal{A}} \; F_{\mu}(h)$',fontsize=24)
plt.xlabel('h',fontsize=24)
plt.title(name,fontsize=24)

#plot the curve
plt.plot(x,y,marker='.')

 #export file for latex, put png to save as image
plt.savefig('./saved/'+name+'.pgf')
plt.savefig('./saved/'+name+'.png',dpi=300)

 #export data as csv
import pandas as pd
d={'1-Threshold':x,'2-Probability of failure':y}
my_df = pd.DataFrame(data=d)
my_df.to_csv('./saved/'+name+'.csv', index=False, header=False,sep=' ')
print(my_df)
