#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 26 14:02:49 2018
Optimisation of the cost function using a differential evolution solver.

@author: Jerome Stenger
"""

from __future__ import division, print_function

import os
import sys
import time
sys.path.append(os.path.join(os.path.dirname(__file__)))

from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE,SIG_DFL) 

from math import *
import numpy as np
import matplotlib.pyplot as plt
from Model import my_model

from mystic.math.discrete import product_measure
from mystic.math import almostEqual

#%%
# =============================================================================
# =============================================================================
############################## SOLVER PARAMETERS ##############################
# =============================================================================
# =============================================================================


MINMAX = 1  ## NOTE: sup = maximize = -1; inf = minimize = 1
npop = 60 # size of the overall population
maxiter = 1000 # maximum number of population mutation
maxfun = 1e+6 
convergence_tol = 1e-6; ngen = 100 # calculation stops with precision equal to tolerance after ngen iteration
crossover = 0.9 # probability of mutation 
percent_change = 0.9 # Scaling factor
#Xinit=[170,3400,13,45,49.5,50.5,54.1,54.9]

def optimize(cost,_bounds,_constraints):
	from mystic.solvers import DifferentialEvolutionSolver2
	from mystic.termination import ChangeOverGeneration as COG
	from mystic.strategy import Best1Exp
	from mystic.monitors import VerboseMonitor, Monitor

	stepmon = VerboseMonitor(5) # step for showing live update
	evalmon = Monitor()
	lb,ub = _bounds
	ndim = len(lb)

	solver = DifferentialEvolutionSolver2(ndim,npop)
	solver.SetRandomInitialPoints(min=lb,max=ub)
	#solver.SetInitialPoints(Xinit,radius=0.9)
	solver.SetStrictRanges(min=lb,max=ub)
	solver.SetEvaluationLimits(maxiter,maxfun)
	solver.SetEvaluationMonitor(evalmon)
	solver.SetGenerationMonitor(stepmon)
	solver.SetConstraints(_constraints)

	tol = convergence_tol
	solver.Solve(cost,termination=COG(tol,ngen),strategy=Best1Exp,CrossProbability=crossover,ScalingFactor=percent_change)

	solved = solver.bestSolution
	func_max = MINMAX * solver.bestEnergy
	func_evals = solver.evaluations
	from mystic.munge import write_support_file
	write_support_file(stepmon, npts=npts)
	return solved, func_max, func_evals



#%%
# =============================================================================
# =============================================================================
####################### O.U.Q WITH THE MYSTIC FRAMEWORK #######################
# =============================================================================
# =============================================================================

#name of the file
name='Mystic Framework'

#threshold bounds
threshold_max = 10
threshold_min = 5
step = 1

ndim=1
npts = (2,)*ndim #Number of Dirac mass per variable, must be lower or equal to the number of constraints + 1.
error=(200,)*ndim #amissible error to the mean

lower=[160]
upper=[3580]

target_mean=list((np.asarray(upper)+np.asarray(lower))/2.) # You can put any value want in a list
target_var=list((np.asarray(upper)-np.asarray(lower))/2.) # Only needed if you want to enforce variance


def costMystic(x):
	global threshold
	x=list(x)
	return my_model(x)-threshold

def constraints(param):
	prodmeasure=product_measure()
	prodmeasure.load(param,npts)
	
	for measure in prodmeasure:
		if not almostEqual(float(measure.mass),1.0):
			measure.normalize()
	
	for i in range(len(prodmeasure)):
		E = float(prodmeasure[i].mean)  # mean
		if not (E <= float(target_mean[i] + error[i])) \
		or not (float(target_mean[i] - error[i]) <= E):
			prodmeasure[i].mean = target_mean[i]
	
# =============================================================================
########################### VARIANCE ENFORCEMENT ##############################
#
#	for i in range(len(prodmeasure)):
# 		E = float(prodmeasure[i].var)  # variance
# 		if not (E <= float(target_var[i] + error[i])) \
# 		or not (float(target_var[i] - error[i]) <= E):
# 			prodmeasure[i].var = target_var[i]
# =============================================================================
	
	return prodmeasure.flatten()


def cost(param):
	prodmeasure = product_measure()
	prodmeasure.load(param, npts)
	
	for i in range(len(prodmeasure)):
		E = float(prodmeasure[i].mean)
		if E > (target_mean[i] + error[i]) or E < (target_mean[i] - error[i]):
			if debug: print("skipping expect: %s" % E)
			return inf  #XXX: FORCE TO SATISFY E CONSTRAINTS MEAN
		
# =============================================================================
########################### VARIANCE ENFORCEMENT ##############################
#
# 	for i in range(len(prodmeasure)):
# 		E = float(prodmeasure[i].var)
# 		if E > (target_var[i] + error[i]) or E < (target_var[i] - error[i]):
# 			if debug: print("skipping expect: %s" % E)
# 			return inf  #XXX: FORCE TO SATISFY E CONSTRAINTS VARIANCE
# =============================================================================
	
	return MINMAX*prodmeasure.pof(costMystic)

# live view
threshold=threshold_max
x=[]
y=[]
plt.rc('text', usetex=True)
plt.rc('font', family='serif')
plt.rc('xtick', labelsize=20) 
plt.rc('ytick', labelsize=20) 
plt.ylabel(r'$\inf_{\mu\in\mathcal{A}} \; F_{\mu}(h)$',fontsize=24)
plt.xlabel('h',fontsize=24)
plt.title(name,fontsize=24)
plt.show()
axes = plt.gca()
axes.set_xlim(threshold_min,threshold_max) # x-axis bounds
axes.set_ylim(0,1) # y-axis bounds
line, = axes.plot(x, y, 'r-')
lb=[]
ub=[]
for i in range(ndim): #specific format to be implemented
	lb=lb+(npts[i]*[0]+npts[i]*[lower[i]])
	ub=ub+(npts[i]*[1]+npts[i]*[upper[i]])

start = time.time()
while threshold>=threshold_min:
	solved, func_max, func_evals = optimize(cost,(lb,ub),constraints)
	c = product_measure()
	c.load(solved,npts)
	#print solution
	for i in range(ndim):
		print("solved variable ",i+1, ' : mean = ', str( round(c[i].mean,4) ),'\n')
		for k in range(npts[i]):#Acceptable_Set_Variance
			print(list(zip([round(c[i].weights[j],4) for j in range(len(c[i].weights))],[round(c[i].positions[j],4) for j in range(len(c[i].positions))]))[k],'\n')
			#	print("variance ", i+1, " : ", str( round(c[i].var,4) ),'\n')   # var(h)
	x.append(threshold)
	y.append(func_max)
	line.set_xdata(x) # add new points to the figure
	line.set_ydata(y) 
	plt.draw() #update the figure
	plt.pause(1e-17)
	threshold+=-step
plt.show()
end = time.time()
print('time elapsed :',end - start)
time.sleep(0.5)

#%%
# =============================================================================
# =============================================================================
################### RESULT VISUALISATION AND EXPORTATION ######################
# =============================================================================
# =============================================================================
plt.close('all')

#Print and saved fig to be printed with latex 
plt.figure(figsize=(12,7)) #figure size
plt.rc('text', usetex=True)
plt.rc('font', family='serif')
plt.rc('xtick', labelsize=20) 
plt.rc('ytick', labelsize=20) 
plt.ylabel(r'$\inf_{\mu\in\mathcal{A}} \; F_{\mu}(h)$',fontsize=24)
plt.xlabel('h',fontsize=24)
plt.title(name,fontsize=24)

#plot the curve
plt.plot(x,y,marker='.')

 #export file for latex, put png to save as image
plt.savefig('./saved/'+name+'.pgf')
plt.savefig('./saved/'+name+'.png',dpi=300)

 #export data, compilation needed to latex with pgdplot, allows to save data
import pandas as pd
d={'1-Threshold':x,'2-Probability of failure':y}
my_df = pd.DataFrame(data=d)
my_df.to_csv('./saved/'+name+'.csv', index=False, header=False,sep=' ')
print(my_df)

