#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 17 13:23:23 2018
Contain our personal implementation of a simulated annealing solver. 4 schedules are available.

@author: Jerome Stenger
"""
from __future__ import print_function
import numpy as np

# =============================================================================
# Fast, Cauchy and Boltzman schedules are adapted from the former simulated annealing algorithm provided in Numpy and now depreciated
# Exp schedule is an adaptation of the Matlab default settings
# =============================================================================
def Fast_SA_ObO(x0,k,T0,lower,upper):#Fast schedule
	T=T0*np.exp(-k)
	u=np.random.uniform(0,1)
	xc=np.sign(u-0.5)*T*((1+1./T)**abs(2*u-1)-1.0)*(upper-lower)
	x_new=max(min(x0+xc,upper),lower)
	return float(x_new),T

def Cauchy_SA_ObO(x0,k,T0,lower,upper):#Cauchy schedule
	T=float(T0)/(1+k)
	nb=np.squeeze(np.random.uniform(-np.pi/2,np.pi/2))
	xc=0.5*T*np.tan(nb)
	x_new=max(min(x0+xc,upper),lower)
	return float(x_new),T

def Boltzmann_SA_ObO(x0,k,T0,lower,upper):#Boltzmann schedule
	T=float(T0)/np.log(1+k)
	std=np.minimum(np.sqrt(T),(upper-lower)/3/0.5)*upper-lower
	xc=np.squeeze(np.random.normal(0,1))
	x_new=max(min(x0+xc*std*0.5,upper),lower)
	return float(x_new),T

def Exp_SA_ObO(x0,k,T0,lower,upper):#Exponential schedule for the temperature decay, Boltzmann move for the exploration.
	T=T0*0.95**k
	std=np.minimum(np.sqrt(T),(upper-lower)/3./0.5)
	xc=np.squeeze(np.random.normal(0,1))
	x_new=max(min(x0+xc*std*0.5,upper),lower)
	return float(x_new),T



def Simulated_Annealing(func,x0,**kwargs):
	lower = kwargs.get('lower', None)
	if not(np.any(lower)):
		lower = len(x0)*[0] #default lower bound
		lower=np.asarray(lower)
	upper = kwargs.get('upper', None)
	if not(np.any(upper)):
		upper = len(x0)*[1] #default upper bound
		upper=np.asarray(upper)
	T0 = kwargs.get('T0',None)
	if not(T0):
		T0=100 # default initial temperature
	schedule = kwargs.get('schedule',None)
	if not(schedule):
		schedule='Exp' # default schedule
	maxiter = kwargs.get('maxiter',None)
	if not(maxiter):
		maxiter=100 # default maxiter, it correspond to the number of re-annealing done
	niter = kwargs.get('niter',None)
	if not(niter):
		niter=10 # default niter, number of new points accepted before re-annealing
		
	x0 =np.asarray(x0) #initial point
	T=T0 # initial temperature, set to 100 if not information are available
	x_best=x0 # best point
	f_best=func(x0) # current minimum
	f_current=func(x0) # current evaluated function
	x_current=x0 # current evaluated point
	x_new=x0 # new evaluated point
	
	
	for j in range(maxiter):
		na=0 # number of accepted solution
		k=1 
		while na<niter:
			for i in range(len(x0)):
				x_new[i],T=eval(schedule+'_SA_ObO(x_current[i],k,T0,lower[i],upper[i])') #evaluation regarding choice of schedule
			f_new = func(x_new)
			dE=(f_new-f_current)
			accept=False
			if dE<0: #always accept new point if value is lower
				accept=True
				if f_new<f_best:
					f_best=f_new #save best value
					x_best=x_new
			else :
				p=1/(1+np.exp(dE/float(T))) #Probability of acceptance of the new point is higher
				if p>np.random.uniform(0,1):
					accept=True
			if accept==True:
				x_current=x_new
				f_current=f_new
				na+=1
			k+=1
		
	print("Minimum at : ", '\n',x_best,'\n',"Minimum value : ", round(f_best,4),sep="")
	return x_best,f_best